/**
 * Chat/console server sent plugin, made by Lorgan3.
 */

// Define global variables
global.chatSendBuffer = buffer_create();
global.chatPacketID = argument1;
global.chatGID = 1;
global.consoleFont = font_add("Lucida Console", 8, 0, 0, 32, 127);

//Define console commands
//Other plugins may have created the map before the plugin is executed
if !variable_global_exists("commandMap") {  //all server commands go here
    global.commandMap = ds_map_create();
}
if !variable_global_exists("voteCommandMap") {  //additional code that gets executed after voting
    global.voteCommandMap = ds_map_create();
}
if !variable_global_exists("clientCommandMap") {   //all client commands go here
    global.clientCommandMap = ds_map_create();
}
if !variable_global_exists("helpMap") { //a string on how to use a specific command
    global.helpMap = ds_map_create();
}

//load standard commands
execute_file(argument0 + "/commands.gml");

//build a list of votable maps
global.votableMaps = ds_list_create();
if file_exists("VotableMaps.txt") {
    var fileHandle, i, mapname;
    fileHandle = file_text_open_read("VotableMaps.txt");
    for(i = 1; !file_text_eof(fileHandle); i += 1) {
        mapname = file_text_read_string(fileHandle);
        // remove leading whitespace from the string
        while(string_char_at(mapname, 0) == " " || string_char_at(mapname, 0) == chr(9)) { // while it starts with a space or tab
          mapname = string_delete(mapname, 0, 1); // delete that space or tab
        }
        if(mapname != "" && string_char_at(mapname, 0) != "#") { // if it's not blank and it's not a comment (starting with #)
            ds_list_add(global.votableMaps, mapname);
        }
        file_text_readln(fileHandle);
    }
    file_text_close(fileHandle);
} else {
    ds_list_copy(global.votableMaps, global.map_rotation);
}

ini_open("gg2.ini");
global.chatGlobal = ini_read_real("Plugins","chat_global",ord('Y'));
global.chatTeam = ini_read_real("Plugins","chat_team",ord('U'));
global.chatToggle = ini_read_real("Plugins","chat_toggle",ord('T'));
global.playerInfo = ini_read_real("Plugins","chat_player_info",1);

// Add some text to the window.
global.chatLog = ds_list_create();
ds_list_add(global.chatLog,"OLorgan3's chat plugin v6 - press T to show/hide,");
ds_list_add(global.chatLog,"OY to chat and U to teamchat.");
ds_list_add(global.chatLog,"OType '/help' for console commands.");
ds_list_add(global.chatLog,"O___________________________________________________");

global.chatTime = ds_list_create();
ds_list_add(global.chatTime,current_time);
ds_list_add(global.chatTime,current_time);
ds_list_add(global.chatTime,current_time);
ds_list_add(global.chatTime,current_time);

// A server host can disable specific commands in the ini.
global.blockedCommands = ds_list_create();
var i, tmp;
tmp = ds_map_find_first(global.commandMap);
if (ini_read_real("Plugins", 'chat_blocked_' + string(tmp), 0) == 1) {
    ds_list_add(global.blockedCommands,tmp);
}

for(i=1; i < ds_map_size(global.commandMap) - 1; i += 1) {
    tmp = string(ds_map_find_next(global.commandMap,tmp));
    if (ini_read_real("Plugins", 'chat_blocked_' + tmp, 0) == 1) {
        ds_list_add(global.blockedCommands,tmp);
    }
}
ini_close();

// Make a new menu for plugin options.
if !variable_global_exists("chatOptions") {
    global.chatOptions = object_add();
    object_set_parent(global.chatOptions, OptionsController);
    object_set_depth(global.chatOptions, -130000);
    object_event_add(global.chatOptions, ev_create, 0, '
        menu_create(40, 140, 300, 200, 30);

        if room != Options {
            menu_setdimmed();
        }

        menu_addback("Back", "
            instance_destroy();
            if(room == Options)
                instance_create(0,0,MainMenuController);
            else
                instance_create(0,0,InGameMenuController);
        ");
    ');

    object_event_add(InGameMenuController, ev_create, 0, '
        menu_addlink("Chat Options", "
            instance_destroy();
            instance_create(0,0,global.chatOptions);
        ");
    ');
}

object_event_add(global.chatOptions, ev_create, 0, '
    menu_addedit_key("Global chat", "global.chatGlobal","");
    menu_addedit_key("Team chat", "global.chatTeam","");
    menu_addedit_key("Toggle chatwindow", "global.chatToggle","");
    menu_addedit_boolean("Display player info:", "global.playerInfo", "");
');

object_event_add(global.chatOptions, ev_destroy, 0, '
    ini_open("gg2.ini");
    ini_write_real("Plugins", "chat_global", global.chatGlobal);
    ini_write_real("Plugins", "chat_team", global.chatTeam);
    ini_write_real("Plugins", "chat_toggle", global.chatToggle);
    ini_write_real("Plugins", "chat_player_info", global.playerInfo);
    ini_close();
');

// Create the new object
object_event_add(PlayerControl, ev_step, ev_step_begin, '
    if (!instance_exists(global.chatWindow)) {
        instance_create(0,0,global.chatWindow);
    }
');

// To prevent errors.
object_event_add(Player, ev_create, 0, '
        hasChat = false;
        nameBkp = "";
        lastChatTime = 0;
        chatCount = 0;
        canVote = false;
        mute = false;
        mayVote = true;
        lastVoteCall = 0;

        gid = global.chatGID;
        global.chatGID += 1;
');

// Because server-sent plugins do weird things on startup + name change detection
object_event_add(Player, ev_step, ev_step_begin, '
    // this only gets executed for host.
    if (!variable_local_exists("hasChat")) {
        hasChat = false;
        nameBkp = "";
        lastChatTime = 0;
        chatCount = 0;
        canVote = false;
        mute = false;
        mayVote = true;
        lastVoteCall = 0;

        gid = global.chatGID;
        global.chatGID += 1;
    }

    if (global.isHost) {
        if (!variable_local_exists("checkIfKicked")) {
            checkIfKicked = true;
            if (instance_exists("global.chatWindow")) {
               if (ds_list_find_index(global.chatWindow.chatKicklist, socket_remote_ip(socket)) != -1) {
                    kicked = true;
                }
            }
        }

        if (name != nameBkp || name == "") {
            if (name == "") {
                name = "Nameless";
                write_ubyte(global.sendBuffer, PLAYER_CHANGENAME);
                write_ubyte(global.sendBuffer, ds_list_find_index(global.players, id));
                write_ubyte(global.sendBuffer, string_length(name));
                write_string(global.sendBuffer, name);
            }

            if (nameBkp == "") {
                nameBkp = name;
                exit;
            }

            if (hasChat || global.myself == id) {
                with(global.chatWindow) {
                    var msg;
                    msg = "O" + other.nameBkp + " has changed their name to " + other.name + ".";
                    execute_string(global.sendChatMessage, msg, 254, 0);
                    execute_string(global.addChatMessage, msg, 254, 0);
                }
                nameBkp = name;
            }
        }
    }
');

// Send a message when someone leaves
object_event_add(Player,ev_destroy, 0, '
    if (hasChat && global.isHost) {
        with(global.chatWindow) {
            var msg;
            msg = "O" + other.name + " has left the game.";
            execute_string(global.sendChatMessage, msg, 254, 0);
            execute_string(global.addChatMessage, msg, 254, 0);
        }
    }
');

// Send a message to the server to tell that we have the chat plugin
object_event_add(PlayerControl, ev_create, 0, '
    if (!global.isHost) {
        write_ubyte(global.chatSendBuffer, 0); //hello
        PluginPacketSend(global.chatPacketID, global.chatSendBuffer);
        buffer_clear(global.chatSendBuffer);
    }
');

// Add the new object and set depth
if (!variable_global_exists('chatWindow')) {
    global.chatWindow = object_add(); //Other plugins may have run before and already injected code.
}

object_set_depth(global.chatWindow, -110002);

// Initialize
object_event_add(global.chatWindow, ev_create, 0, '
    open = false;
    hidden = true;
    team = false;
    offset = 0;
    image_speed = 0;

    votePlayer = noone;
    votes = ds_list_create();
    voteTimeout = 10000; // miliseconds

    options = ds_list_create();
    voteCommand = ds_list_create();

    selectedmap = false;

    fakeMaprotation = ds_list_create(); // for mapvoting
    oldRotation = global.map_rotation;
    oldMapIndex = -1;

    chatBanlist = ds_list_create();
    chatKicklist = ds_list_create();
');

// Alarm 0 happens when a vote ends
object_event_add(global.chatWindow, ev_alarm, 0, '
    with(Player) {
        mayVote = true; // Reset this var after the vote is over
    }

    if (global.isHost && votePlayer == noone) {
        if (ds_list_size(voteCommand) > 0) {
            execute_string(ds_map_find_value(global.voteCommandMap, ds_list_find_value(voteCommand, 0)), voteCommand, options, votes, votePlayer);
        } else {
            with(global.voter) {
                instance_destroy();
            }
        }
    }

    ds_list_clear(voteCommand);
    ds_list_clear(votes);
    ds_list_clear(options);
    votePlayer = noone;
');

// All main stuff happens here
object_event_add(global.chatWindow, ev_step, ev_step_end, '
    x = view_xview[0] + 7;
    y = view_yview[0] + view_hview[0] - 243;
    randomColor = make_color_rgb(random(255), random(255), random(255));

    if (!instance_exists(InGameMenuController) &&
        !instance_exists(OptionsController) &&
        !instance_exists(ControlsController) &&
        !instance_exists(HUDOptionsController)) ||
        instance_exists(global.cheatycheat) {

        if (keyboard_check_pressed(global.chatToggle) && !open) {
            if (!open) {
                hidden = !hidden;
            }
        } else if (keyboard_check_pressed(global.chatGlobal) && !open) {
            open = true;
            keyboard_string = "";
            team = false;
            instance_create(0, 0, global.cheatycheat);
        } else if (keyboard_check_pressed(global.chatTeam) && global.myself.team != TEAM_SPECTATOR && !open) {
            open = true;
            keyboard_string = "";
            team = true;
            instance_create(0, 0, global.cheatycheat);
        } else if (open && keyboard_check_pressed(vk_enter)) {
            with(global.cheatycheat) {
                instance_destroy();
            }

            if (keyboard_string != "") {
                if (global.isHost) {
                    var _team;
                    if (team == true) {
                        _team = global.myself.team+1;
                    } else {
                        _team = 0;
                    }
                    execute_string(global.executeChat, keyboard_string, global.myself, _team)
                } else {
                    _team = team;
                    if (string_char_at(keyboard_string, 1) == "/") {
                        var arguments;
                        arguments = execute_string(global.parseCommand, keyboard_string);
                        if (ds_map_exists(global.clientCommandMap, string_lower(ds_list_find_value(arguments, 0)))) {
                            execute_string(ds_map_find_value(global.clientCommandMap, string_lower(ds_list_find_value(arguments, 0))), arguments);
                        } else {
                            execute_string(global.sendChatMessageClient, keyboard_string, team);
                        }

                        ds_list_destroy(arguments);
                    } else {
                        execute_string(global.sendChatMessageClient, keyboard_string, team);
                    }
                }

                keyboard_string = "";
            }

            open = false;
            team = false;
        }
    }

    if (open) {
        image_index = 1;
        if ((keyboard_check_pressed(vk_up) or mouse_wheel_up()) && ds_list_size(global.chatLog) > 9) {
            offset = min(ds_list_size(global.chatLog) - 10, offset + 1);
        } else if (keyboard_check_pressed(vk_down) or mouse_wheel_down()) {
            offset = max(0, offset - 1);
        }
    } else {
        image_index = 0;
    }

    if global.isHost {
        event_user(0); //forward messages as a host
    } else {
        event_user(1); //read messages as a client
    }
');

//draw
object_event_add(global.chatWindow, ev_draw, 0, '
    var xoffset, yoffset, xsize, ysize, index;
    xsize = 380;
    ysize = 180;

    draw_set_valign(fa_top);
    draw_set_halign(fa_left);
    draw_set_font(global.consoleFont);

    if (!hidden || open) {
        draw_set_color(make_color_rgb(61, 61, 61));
        draw_set_alpha(0.85);
        draw_rectangle(x, y, x + xsize, y + 155, false);
        if (open) {
            draw_rectangle(x, y + 160, x + xsize, y + ysize, false);
        }

        draw_set_color(c_black);
        draw_set_alpha(1);
        draw_rectangle(x, y, x + xsize, y + 155, true);
        if (open) {
            draw_rectangle(x, y + 160, x + xsize, y + ysize, true);
        }
    } else {
        draw_set_alpha(1);
    }

    draw_set_color(c_white);

    //prevent too long strings
    if (open) {
        keyboard_string = string_copy(keyboard_string, 1, 255); //limit messages

        // Make that the text doesnt go outside the chatwindow
        if (string_length(keyboard_string) - 53 > 0) {
            index  = string_length(keyboard_string) - 52;
        } else {
            index = 0;
        }
        chatmessage = string_copy(keyboard_string, index, 53);
        if (team == 1) {
            if (global.myself.team == TEAM_BLUE) {
                colorCode = "B";
                event_user(2);
            } else {
                colorCode = "R";
                event_user(2);
            }
        } else if (string_char_at(keyboard_string, 1) == "/") {
            colorCode = "O";
            event_user(2);
        }
        draw_text(x + 3, y + ysize - 15, chatmessage);
    }

    // The drawing of the text
    var amount;
    if (!hidden || open) {
        amount = min(ds_list_size(global.chatLog), 10);
    } else {
        for(i = ds_list_size(global.chatLog) - 1; i >= ds_list_size(global.chatLog) - 10 && i >= 0; i -= 1) {
            if (current_time-ds_list_find_value(global.chatTime, i) > 10000) {
                break;
            }
        }
        amount = ds_list_size(global.chatLog) - 1 - i;
    }

    for(i=0; i < amount; i += 1) {
        var msg, prefix, pos;
        msg = ds_list_find_value(global.chatLog, ds_list_size(global.chatLog) - amount + i - offset);
        prefix = "";
        pos = string_pos(chr(13), msg);
        if (pos != 0) {
            colorCode = string_copy(msg, 1, 1);
            event_user(2);

            prefix = string_copy(msg, 2, pos - 1);
            draw_text(x + 10, y + ysize - 35 - 13 * (amount - i), prefix);
            msg = string_copy(msg, pos + 1, 255);
        }

        colorCode = string_copy(msg, 1, 1);
        event_user(2);
        msg = string_copy(msg, 2, 255);
        draw_text(x + 10 + string_width(prefix), y + ysize - 35 - 13 * (amount - i), msg);
    }

    draw_set_font(global.gg2Font);
');

//event0 : process input as a host
object_event_add(global.chatWindow, ev_other, ev_user0, '
    while(PluginPacketGetBuffer(global.chatPacketID) != -1) {
        var player;
        chatReceiveBuffer = PluginPacketGetBuffer(global.chatPacketID);
        player = PluginPacketGetPlayer(global.chatPacketID);
        if (player == -1 || !instance_exists(player)) {
            continue;
        }

        switch(read_ubyte(chatReceiveBuffer)) {
            case 0: //hello
                if (ds_list_find_index(chatBanlist,socket_remote_ip(player.socket)) == -1) {  //ignore muted players
                    player.hasChat = true;
                    player.nameBkp = player.name;

                    msg = "O" + player.name + " has joined chat!";
                    execute_string(global.sendChatMessage, msg, 254, 0);
                    execute_string(global.addChatMessage, msg, 254, 0);

                    // Sync votes if needed
                    with(global.voter) {
                        if (win == -1) {
                            write_ubyte(global.chatSendBuffer, 204);
                            write_ubyte(global.chatSendBuffer, min(255, string_length(question)));
                            write_string(global.chatSendBuffer, string_copy(question, 0, 255));
                            write_ubyte(global.chatSendBuffer, ds_list_size(options));

                            var i;
                            for(i=0; i < ds_list_size(options); i += 1) {
                                write_ubyte(global.chatSendBuffer, min(255, string_length(ds_list_find_value(options, i))));
                                write_string(global.chatSendBuffer, string_copy(ds_list_find_value(options, i), 0, 255));
                            }

                            write_ushort(global.chatSendBuffer,ceil(alarm[0] * global.delta_factor));

                            for(i=0; i<ds_list_size(options); i+=1) {
                                write_ubyte(global.chatSendBuffer, optionVotes[i]);
                            }

                            with(player) {
                                canVote = true;
                                PluginPacketSendTo(global.chatPacketID, global.chatSendBuffer, id);
                            }

                            buffer_clear(global.chatSendBuffer);
                        }
                    }
                }
            break;
            case 1: //global chat
                if (player.hasChat) {
                    var len, msg;
                    len = read_ubyte(chatReceiveBuffer);
                    msg = read_string(chatReceiveBuffer, len);
                    execute_string(global.executeChat, msg, player, 0);
                }
            break;
            case 2: //team chat
                if (player.team != TEAM_SPECTATOR && player.hasChat) {
                    var team, len, msg, sender;
                    team = 1 + player.team;
                    len = read_ubyte(chatReceiveBuffer);
                    msg = read_string(chatReceiveBuffer, len);
                    sender = ds_list_find_index(global.players, player);

                    execute_string(global.sendChatMessage, msg, sender, team);
                    if (global.myself.team == TEAM_SPECTATOR || team == global.myself.team + 1) {
                        execute_string(global.addChatMessage, msg, sender, team);
                    }
                }
            break;
            case 3: //vote reply
                var vote;
                vote = read_ubyte(chatReceiveBuffer);
                if (player.canVote && player.mayVote) {
                    player.canVote = false;
                    if (vote <= 9) {
                        execute_string(global.handleVote, votes, voteCommand, options, votePlayer, player, vote);
                    }
                }
            break;
        }

        buffer_clear(chatReceiveBuffer);
        PluginPacketPop(global.chatPacketID);
    }
');


//event1 : read messages as a client
object_event_add(global.chatWindow, ev_other, ev_user1, '
    while(PluginPacketGetBuffer(global.chatPacketID) != -1) {
        var playerId, player, team, msg;
        chatReceiveBuffer = PluginPacketGetBuffer(global.chatPacketID);
        playerId = read_ubyte(chatReceiveBuffer);
        if (playerId < 200) {
            team = read_ubyte(chatReceiveBuffer);
            msg = read_string(chatReceiveBuffer, read_ubyte(chatReceiveBuffer));
            execute_string(global.addChatMessage, msg, playerId, team);
        } else if (playerId == 200) {
            with(global.voter) {
                instance_destroy();
            }
            instance_create(0, 0, global.voter);

            with(global.voter) {
                question = read_string(other.chatReceiveBuffer, read_ubyte(other.chatReceiveBuffer));
                var _options, i;
                _options = read_ubyte(other.chatReceiveBuffer);
                ds_list_clear(options);
                for(i=0; i < _options && i < 9; i += 1) {
                    ds_list_add(options, read_string(other.chatReceiveBuffer, read_ubyte(other.chatReceiveBuffer)));
                }

                alarm[0] = read_ushort(other.chatReceiveBuffer) / global.delta_factor;
            }
        } else if (playerId == 201) {
            var _win;
            _win = read_ubyte(chatReceiveBuffer);
            if (_win == 0) {
                with(global.voter) {
                    instance_destroy();
                }
            } else {
                with(global.voter) {
                    win = _win;
                    done = true;
                    alarm[0] = 90 / global.delta_factor;
                }
            }
        } else if (playerId == 203) { // some option has more votes now
            var option;
            option = read_ubyte(chatReceiveBuffer);
            with(global.voter) {
                optionVotes[option] += 1;
                total += 1;
            }
        } else if (playerId == 204) { // this packet contains a vote + statistics for when you join an existing vote
            with(global.voter) {
                instance_destroy();
            }
            instance_create(0, 0, global.voter);

            with(global.voter) {
                question = read_string(other.chatReceiveBuffer, read_ubyte(other.chatReceiveBuffer));
                var _options, i;
                _options = read_ubyte(other.chatReceiveBuffer);
                ds_list_clear(options);
                for(i=0; i < _options && i < 9; i += 1) {
                    ds_list_add(options, read_string(other.chatReceiveBuffer, read_ubyte(other.chatReceiveBuffer)));
                }

                alarm[0] = read_ushort(other.chatReceiveBuffer) / global.delta_factor;

                for(i=0; i < _options; i += 1) {
                    optionVotes[i] = read_ubyte(other.chatReceiveBuffer);
                    total += optionVotes[i];
                }
            }
        } else {
            msg = read_string(chatReceiveBuffer, read_ubyte(chatReceiveBuffer));
            execute_string(global.addChatMessage, msg, playerId, 0);
        }

        buffer_clear(chatReceiveBuffer);
        PluginPacketPop(global.chatPacketID);
    }
');

//event2 : set color
object_event_add(global.chatWindow, ev_other, ev_user2, '
    switch(colorCode) {
        case "R":
            draw_set_color(make_color_rgb(237, 61, 61));
            break;
        case "B":
            draw_set_color(make_color_rgb(0, 128, 255));
            break;
        case "G":
            draw_set_color(make_color_rgb(32, 183, 28));
            break;
        case "O":
            draw_set_color(c_orange);
            break;
        case "A":
            draw_set_color(c_aqua);
            break;
        case "S":
            draw_set_color(global.chatWindow.randomColor);
            break;
        default:
            draw_set_color(c_white);
    }
');

object_event_add(global.chatWindow, ev_destroy, 0, '
    ds_list_destroy(votes);
    ds_list_destroy(options);
    ds_list_destroy(voteCommand);

    if global.map_rotation == fakeMaprotation {
        global.map_rotation = oldRotation;
        GameServer.currentMapIndex = oldMapIndex;
    }
    ds_list_destroy(fakeMaprotation)

    ds_list_destroy(chatBanlist);
    ds_list_destroy(chatKicklist);
');

// Import the other objects
execute_file(argument0 + '/vote_api.gml');
execute_file(argument0 + '/chat_api.gml');
execute_file(argument0 + '/fake_menu.gml');

execute_file(argument0 + '/lorganbase.gml');
ds_list_add(lorganbase_plugins, 'chat');
